from setuptools import setup, find_packages
setup(name="mwcommon",
      version="0.1",
      py_modules=["cfgtools", "msgbus", "persistence", "sysactions", "systools", "ui"],
      install_requires=[],
      url="http://www.e-deploy.com.br",
      author="Eduardo Almeida",
      author_email="eduardo.almeida@e-deploy.com.br")
